#include <stdarg.h>
#include <stddef.h>
#include <setjmp.h>
#include <cmocka.h>

#include <libubus.h>
#include <libubox/blobmsg_json.h>
#include <libubox/blobmsg.h>
#include <json-c/json_tokener.h>
#include <json-validator.h>
#include <json-c/json.h>
#include <json-editor.h>
#include <uci.h>

#include "network.h"
#include "config.h"
#include "network_tests.h"
#include "tools.h"

#define _DEBUG_ 1
#define DEBUG(...) \
	if (_DEBUG_)   \
	printf(__VA_ARGS__)

#define QD_LINE_MAX 2048

static char *msg_sent = NULL;
void populate_hosts_from_clients(void);
void populate_clients(void);
void get_bridge_info(Client *clnt);
/* Forward declarations */

int quest_router_hosts(struct ubus_context *ctx, struct ubus_object *obj,
					   struct ubus_request_data *req, const char *method,
					   struct blob_attr *msg);

int quest_network_reload(struct ubus_context *ctx, struct ubus_object *obj,
						 struct ubus_request_data *req, const char *method,
						 struct blob_attr *msg);

int network_init(void);
int network_exit(void);

static void reset_msg_sent(void)
{
	if (msg_sent)
	{
		free(msg_sent);
		msg_sent = NULL;
	}
}

/* Override ubus call */
int ubus_send_reply(struct ubus_context *ctx, struct ubus_request_data *req,
					struct blob_attr *msg)
{
	reset_msg_sent();
	if (msg)
		msg_sent = blobmsg_format_json(msg, true);
	return 0;
}

/*Override questd add object */
int quest_add_object(struct ubus_object *obj)
{
	return 0;
}
/*Override monitor clients */
void *monitor_clients(void *arg)
{
	populate_clients();
	populate_hosts_from_clients();
	return NULL;
}
/*Override bridge info */
void get_bridge_info(Client *clnt)
{
	clnt->brportno = 1;
	clnt->connected = true;
}
static void test_router_hosts(void)
{
	struct json_object *jobj, *hosts, *host, *device, *macaddr;
	int two_iface = 0, five_iface = 0;
	int arraylen, i;

	sleep(2);
	assert_true(quest_router_hosts(NULL, NULL, NULL, NULL, NULL) == 0);
	assert_non_null(msg_sent);
	jobj = json_tokener_parse(msg_sent);

	hosts = json_object_get_by_string(jobj, "hosts");
	arraylen = json_object_array_length(hosts);
	for (i = 0; i < arraylen; i++)
	{
		host = json_object_array_get_idx(hosts, i);
		device = json_object_get_by_string(host, "device");
		macaddr = json_object_get_by_string(host, "macaddr");
		if (!strcmp(FIVE_IFACE, json_object_get_string(device)) && !strcmp(FIVE_CLIENT_MAC, json_object_get_string(macaddr)))
			five_iface = 1;
		if (!strcmp(TWO_IFACE, json_object_get_string(device)) && !strcmp(TWO_CLIENT_MAC, json_object_get_string(macaddr)))
			two_iface = 1;
	}
	assert_true(two_iface & five_iface);
	json_object_put(jobj);
	DEBUG("hosts:\n%s\n", msg_sent);
}


static void test_router_reload(void)
{
	struct uci_element *e;
	struct uci_context *uci_ctx = NULL;
	struct uci_package *uci_network;
	const char *before[8] = {0};
	const char *is_lan_before = NULL;
	const char *is_lan = NULL;
	const char *is_lan_after = NULL;
	const char after[8] = {0};

	uci_ctx = uci_alloc_context();
	uci_network = init_package(&uci_ctx, "network");
	assert_non_null(uci_network);

	assert_true(quest_router_networks(NULL, NULL, NULL, NULL, NULL) == 0);
	DEBUG("reload:\n%s\n", msg_sent);

	uci_foreach_element(&uci_network->sections, e) {
		struct uci_section *s = uci_to_section(e);

		if (strcmp(s->type, "interface") != 0)
			continue;

		is_lan_before = uci_lookup_option_string(uci_ctx, s, "is_lan");
		if(!is_lan_before)
			continue;

		strcpy(before, is_lan_before);
	}

	assert_string_equal(before, "1");

	uci_foreach_element(&uci_network->sections, e) {
		struct uci_section *s = uci_to_section(e);
		if (strcmp(s->type, "interface") == 0) {
			is_lan = uci_lookup_option_string(uci_ctx, s, "is_lan");
			if(is_lan) {
				struct uci_ptr ptr = {0};
				ptr.p = uci_network;
				ptr.s = s;
				ptr.option = "is_lan";
				ptr.value = "0";
				uci_set(uci_ctx, &ptr);
			}
		}
	}

	uci_commit(uci_ctx, &uci_network, false);

	assert_true(quest_network_reload(NULL, NULL, NULL, NULL, NULL) == 0);
	assert_true(quest_router_networks(NULL, NULL, NULL, NULL, NULL) == 0);

	uci_foreach_element(&uci_network->sections, e) {
		struct uci_section *s = uci_to_section(e);
		if (strcmp(s->type, "interface") != 0)
			continue;

		is_lan_after= uci_lookup_option_string(uci_ctx, s, "is_lan");
		if(!is_lan_after)
			continue;

		strcpy(after, is_lan_after);
	}

	assert_string_equal(after, "0");
	assert_non_null(msg_sent);
	DEBUG("reload:\n%s\n", msg_sent);
	uci_unload(uci_ctx, uci_network);
	uci_free_context(uci_ctx);
}

static int group_setup(void **state)
{
	network_init();
	return 0;
}

static int group_teardown(void **state)
{
	struct uci_element *e;
	struct uci_context *uci_ctx = NULL;
	struct uci_package *uci_network;
	const char *is_lan = NULL;

	uci_ctx = uci_alloc_context();
	uci_network = init_package(&uci_ctx, "network");
	uci_foreach_element(&uci_network->sections, e) {
		struct uci_section *s = uci_to_section(e);
		struct uci_ptr ptr = {0};

		if (strcmp(s->type, "interface") != 0)
			continue;

		is_lan = uci_lookup_option_string(uci_ctx, s, "is_lan");
		if(!is_lan)
			continue;

		ptr.p = uci_network;
		ptr.s = s;
		ptr.option = "is_lan";
		ptr.value = "1";
		uci_set(uci_ctx, &ptr);
	}

	uci_commit(uci_ctx, &uci_network, false);

	uci_unload(uci_ctx, uci_network);
	uci_free_context(uci_ctx);
	reset_msg_sent();
	network_exit();
	return 0;
}

int main(void)
{
	const struct CMUnitTest tests[] = {
		cmocka_unit_test(test_router_hosts),
		cmocka_unit_test(test_router_reload),
	};
	return cmocka_run_group_tests(tests, group_setup, group_teardown);
}